//
// @(#)magarray.cpp<2>  1.1  misha-12nov104
//
// Copyright (c) 2004 of Mikhail Tchernychev.
// All rights reserved.
//
// Created:  12nov104  by  misha@misha.local
// Version:  12nov104  by  misha@misha.local
//


#include <iostream>

#include <stdio.h>
#include <stdlib.h>

#include "magobject.h"
#include "magdipole.h"
#include "magcollection.h"
#include "maggriddata.h"
#include "magprofiles.h"
#include "maginv.h"
#include "magfunc.h"
#include "magarray.h"
#include "magbackground.h"
#include "magbackground2.h"

#include "l2estimator.h"
#include "l1estimator.h"
#include "dnls1estimator.h"
#include "tn_estimator.h"
#include "magchannel.h"

#include "targetarray.h"

using namespace std;


class CLimitedInversion: public CMagInv {
 
public:
    
   double m_lfZmin;
   double m_lfZmax;    
   int m_iCounter;

CLimitedInversion() {    
    m_lfZmin = -1.e+38;
    m_lfZmax = 1.e+38;
    m_iCounter = 0;
};      
    
};    


#include <signal.h>

static int ibreak = 0;

void ctrl_c(int sig) {
  ibreak = 1;
  fprintf(stderr,"signal\n");
}

int progress_func(void * data, int percent, double discr) { // progress / cancel function

  CLimitedInversion * pInversion = (CLimitedInversion *)data;
  double vars[3];

  pInversion->GetNonLinearVariables(vars, 3);

  pInversion->m_iCounter++;
  if(pInversion->m_iCounter < 5) return 0;
  if(vars[2] > pInversion->m_lfZmax || vars[2] < pInversion->m_lfZmin) return 1;  
   
  fprintf(stderr,"iter=%d discrepancy=%lf\n", percent, discr); fflush(stderr);
  return ibreak;

}


main(int argc, char ** argv) {

  signal(SIGINT,ctrl_c);
 
  double positions[3], J[3], max_err, std_err;
 
  double depth   = 2.;
  double I       = 65.6; 
  double D       = -12.27;
  double Az      = 90.;
  int n_channels = 2;
  double zmin    = 100;
  double zmax    = 100; 
  int background = 1;
  double dm      = 1.e-3;
  int is_shift   = 0;
  double mag_ampl = 0.;
  int is_induced = 0;

 
  if(argc == 1) {
    fprintf(stderr,"\nUsage: %s -I val -D val -Ch val -Bg val -Z val -dm val < input > output\n",argv[0]); 
    fprintf(stderr,"-I  number  -  Inclination, degrees\n"); 
    fprintf(stderr,"-D  number  -  Declination, degrees\n"); 
    fprintf(stderr,"-Ch number  - number of channels [2] \n"); 
    fprintf(stderr,"-Z  number  - initial depth below sea bottom, m [2 m]\n"); 
    fprintf(stderr,"-Bg number  - background order (1 or 2) [1]\n"); 
    fprintf(stderr,"-dm number  - smooth depth order [1.e-3]\n"); 
    fprintf(stderr,"-S          - shift all profiles except first\n"); 
    fprintf(stderr,"-Az number  - Local X axis azimuth [90]\n");
    fprintf(stderr,"-Ind        - Induced only\n");             
    exit(0);
  }


for(int i=1; i<argc; i++) {
    if(argv[i][0]=='-') {
      if(strcmp(argv[i],"-I")==0) {
	  I = atof(argv[i+1]); i++; continue;
      }
      if(strcmp(argv[i],"-D")==0) {
 	  D = atof(argv[i+1]); i++;  continue;
      }
      if(strcmp(argv[i],"-Ch")==0) {
   	  n_channels = atol(argv[i+1]); i++; continue;
      }
      if(strcmp(argv[i],"-Z")==0) {
	  depth = atof(argv[i+1]); i++; continue;
      }
      if(strcmp(argv[i],"-Bg")==0) {
	  background = atoi(argv[i+1]); i++; continue;
      }
      if(strcmp(argv[i],"-dm")==0) {
	  dm = atof(argv[i+1]); i++; continue;
	  }	  
      if(strcmp(argv[i],"-S")==0) {
	  is_shift = 1; continue;	  
      }
      if(strcmp(argv[i],"-Az")==0) {
	  Az = atoi(argv[i+1]); i++; continue;
      }
      if(strcmp(argv[i],"-Ind")==0) {
	  is_induced = 1; continue;	  
      }    
      }
  }



  // create objects we need in inversion

  CMagProfileArray *pMagArray = new CMagProfileArray;

  if(!pMagArray) {
      fprintf(stderr,"Can not create pMagArray. Stop.\n");
      exit(0);
  }    

  pMagArray->SetChannels(n_channels);

  CMagDipole * pDipole = new CMagDipole(is_induced);
  
  if(!pDipole) {
      fprintf(stderr,"Can not create pDipole. Stop.\n");
      delete pMagArray;
      exit(0);
  }        
  
  CMagObject *  pBackground;
  
  if(background == 1)
      pBackground = new CMagBackGround;
  else
      pBackground = new CMagBackGround2;
            
  if(!pBackground) {
        fprintf(stderr,"Can not create pBackground. Stop.\n");    
        delete pMagArray;
        delete pDipole;
        exit(0);
    }    
 
  
  // read data
  
  char buffer[1024], date_txt[20], time_txt[20], name[20];
  const char *IFS = " \t\n";
  double x[10], y[10], z[10], alt[10], T[10];
     
  fgets(buffer, 1024, stdin);
  int k = 0; 
  double x_middle, y_middle;
              
          
  while(!feof(stdin)) {
      if(!fgets(buffer, 1024, stdin)) break;
      
      char * p  = strtok (buffer, IFS);
                    
      for(int i=0; i<n_channels; i++) {
          x[i]   =  atof(p); p = strtok(NULL, IFS);
          y[i]   =  atof(p); p = strtok(NULL, IFS);
          z[i]   =  atof(p); p = strtok(NULL, IFS);           
          alt[i] =  atof(p); p = strtok(NULL, IFS);
          T[i]   =  atof(p); p = strtok(NULL, IFS);                         
      }    
      
      double dtime = 0.;
      strcpy(name,p);     p = strtok(NULL, IFS);
      if(p) {
            strcpy(date_txt,p); p = strtok(NULL, IFS);
            strcpy(time_txt,p); 
            dtime = date2sec(date_txt, time_txt, 1);
      }

      if(!k){
          pMagArray->InitProfile(name);
          pMagArray->SetChannels(n_channels);
      }    
          
     pMagArray->Add(x, y, dtime, T, z, alt);
     k++;
   }   
   
   int synt_size = k*n_channels;
   double * T_synt = new double[synt_size];

   if( dm > 0.) pMagArray->DepthSplineSmooth(dm);
           
   // tmp binary file
   
   unsigned long fhandle = 0;
    
    
   CMagDataCollection DataCollection;
   CMagCollection     ObjectCollection;
   
   ObjectCollection.AddMagObject(pDipole);
   ObjectCollection.AddMagObject(pBackground); 
 
   CMagChannel * pChannel2 = NULL;
         
   for(int i=0; i<n_channels; i++) {       
      CMagChannel * pChannel = new CMagChannel;
      pChannel->SetChannel(i, pMagArray); 
      DataCollection.AddData(pChannel);
      if(is_shift && i) {
         pChannel->FixNonLinearParam(X_OFFSET, FALSE, 0.);
         pChannel->FixNonLinearParam(Y_OFFSET, FALSE, 0.);
         ObjectCollection.AddMagObject(pChannel); 
     } 
     if(i==1)  pChannel2 = pChannel;           
   }
    
   CLimitedInversion Inversion;
   Inversion.SetProgressFunc(progress_func);

   Inversion.SetMagData(&DataCollection);   
   Inversion.SetMagCollection(&ObjectCollection);
  
   Inversion.SetEarthFieldDirections(I,D,Az);

   // limits

   CTargetArray targetarray(10., 2., 25., -1.);
   double res;
         
     pMagArray->SetStartStop(0, pMagArray->GetTotalTimePoints()-1);

     //pMagArray->CheckBackgroundL1(&res);
     
     //if(res <= MaxLinear) continue;

    double median_depth = pMagArray->GetMedianDepth();  
    double median_alt   = pMagArray->GetMedianAlt();
  
    double middle_pos[4], x2[3];
    pMagArray->GetMiddlePointCoords(&middle_pos[0], &middle_pos[1],
                                     &middle_pos[2], &middle_pos[3]);
                                                                
      middle_pos[2] = median_depth+median_alt+depth;
      Inversion.m_lfZmin = median_depth-zmin;
      Inversion.m_lfZmax = median_depth+median_alt+zmax;     
      Inversion.m_iCounter = 0;
      x2[0] = x2[1] = x2[3] = 0;
      
      // here inversion
      
     pDipole->SetNonLinearParams(middle_pos, 3);
     int n_total_data, n_linear_vars, n_nonline_vars;
       
     DataCollection.ReScan();
     ObjectCollection.ReScanAllObjects(&n_linear_vars, &n_nonline_vars);

     // linear inversion , no dipole

     pDipole->FixLinearParam(0, TRUE, 0.);
     pDipole->FixLinearParam(1, TRUE, 0.);
     pDipole->FixLinearParam(2, TRUE, 0.);

     Inversion.GetTotalParameters(&n_total_data, &n_linear_vars, &n_nonline_vars);
     
     double fit, linear_fit;
     Z29RLS_PARAMS rls_params;
     rls_params.dm = 1.e+15;
     
     Inversion.EstimateLinearParameters(Estimate_L2_rls, &linear_fit, 1, &rls_params);     
     double max_linear = Inversion.GetLinearMaxDiff(); 
     // non-linear

     pDipole->FixLinearParam(0, FALSE, 0.);
     pDipole->FixLinearParam(1, FALSE, 0.);
     pDipole->FixLinearParam(2, FALSE, 0.);

     Inversion.GetTotalParameters(&n_total_data, &n_linear_vars, &n_nonline_vars);
     
     rls_params.dm = 1.e+15;
     Inversion.EstimateLinearParameters(Estimate_L2_rls, &fit, 1, &rls_params);

     DNLS1_STORAGE dstorage;
     dstorage.iopt =  2;
     dstorage.maxfev= 10000;
     char message[1024];
     Inversion.EstimateAllParameters(Estimator_dnls1, &dstorage, message);
    
     std_err = Inversion.GetNonLinearDiscrepancy();
     max_err = Inversion.GetNonLinearMaxDiff();
    
    Inversion.DumpObjects(stderr);
    
     // compute diploe field only
    if(T_synt) {
        int nb = pBackground->GetAllLinearParams();
        for(int jj=0; jj<nb; jj++) {  
           pBackground->FixLinearParam(jj, TRUE, 0.);
       }
       double mag_min, mag_max;
       mag_min =  1.e+38;
       mag_max = -1.e+38;
       Inversion.ComputeField(T_synt, synt_size);
       for(int jj=0; jj<synt_size; jj++) {
           mag_min = min(mag_min, T_synt[jj]);
           mag_max = max(mag_max, T_synt[jj]);          
       }    
       mag_ampl = fabs(mag_min-mag_max);
    }   
         
  
     if(pChannel2) pChannel2->GetNonLinearParams(x2, 3);
  
     if(dstorage.info != 5) {
	    targetarray.AddTarget(&Inversion, pDipole, pMagArray, fhandle, 
        linear_fit/std_err, (char *)pMagArray->GetName(), x2, background,mag_ampl);
     }

    fprintf(stderr,"linear fit=%lf non_linear %lf %lf\n", linear_fit, std_err, std_err/linear_fit); 
    fprintf(stderr,"linear max fit=%lf non_linear max %lf %lf\n", max_linear, max_err, max_linear/max_err);  
            
   targetarray.PrintAll(stdout, pMagArray->GetTotalTimePoints());  
   pMagArray->Dump(stdout);
 

  delete pMagArray;
  if(T_synt) delete [] T_synt;

} 
