//
// @(#)targetarray.h  1.1  misha-12nov104
//
// Copyright (c) 2004 of Mikhail Tchernychev.
// All rights reserved.
//
// Created:  12nov104  by  misha@misha.local
// Version:  12nov104  by  misha@misha.local
//


#ifndef TARGETARRAY_H
#define TARGETARRAY_H

#include <vector>
#include <string>
#include <map>

#include "magobject.h"
#include "magdipole.h"
#include "magcollection.h"
#include "maggriddata.h"
#include "magprofiles.h"
#include "maginv.h"
#include "magfunc.h"
#include "magarray.h"
#include "magbackground.h"



using namespace std;


typedef struct _target {
    double x[3];
    double J[3];
    double dx[3];
    double x2[3];
    unsigned long file_handle;
    double fit;
    int count;
    double time;
    double depth;
    double alt;
    int n_start;
    int n_stop;
    double Jtotal, incl, decl, angle;
    double ratio;
    char line[20];
    int bg;
    double mag_ampl;

    _target const &operator=(_target const &o); // overloaded assignment

} TARGET;


class CTargetArray {

private:

    vector<TARGET *> m_vTargets;
    double m_lfMaxCrossDistance;
    double m_lfMaxLongDistance;
    double m_lfPercent;
    double m_lfFitTolerance;

    void destroy();

public:

    CTargetArray();
    CTargetArray(double cross, double along, double percent, double tolerance);
    ~CTargetArray();


    int AddTarget(double * x, double *J, double nx, double ny,
		   double x0, double y0, double len,
		   double fit, unsigned long fhandle);

    int AddTarget(CMagInv * pInversion, CMagDipole * pDipole, CMagProfileArray *pMagArray,
		  unsigned long fhandle, double ratio, char * line, double * x2, int bg, double mag_ampl);
    
    int GetTotalTargets() { return m_vTargets.size(); }
    int GetTarget(int n, double *x, double *J, unsigned long *fhandle, double * fit);

    void PrintShortList(FILE * out, FILE * bin_file, FILE * field_out);
    void PrintAll(FILE * dat, int n_data);
};


TARGET * MakeTarget(double *x, double *J, double *dx, unsigned long file_handle,
		    double fit, double time, double depth, double alt,
		    double Jtotal, double incl, double decl, double angle,
		    int n_start, int n_end, double ratio, char * line, double *x2, int bg,
            double mag_ampl);

int DumpBinaryOut(unsigned long fhandle, FILE * in, FILE *out, int n_start);

#endif
