//
// @(#)maggrad.cpp  1.1  misha-02jun105
//
// Copyright (c) 2005 of Mikhail Tchernychev.
// All rights reserved.
//
// Created:  02jun105  by  misha@MISHA
// Version:  02jun105  by  misha@MISHA
//


#include <iostream>

#include <stdio.h>
#include <stdlib.h>

#include "magobject.h"
#include "magdipole.h"
#include "magcollection.h"
#include "maggriddata.h"
#include "magprofiles.h"
#include "gradprofile.h"
#include "maginv.h"
#include "magfunc.h"
#include "magarray.h"
#include "magbackground.h"
#include "magbackground2.h"

#include "l2estimator.h"
#include "l1estimator.h"
#include "dnls1estimator.h"



using namespace std;



#include <signal.h>

static int ibreak = 0;

void ctrl_c(int sig) {
  ibreak = 1;
  fprintf(stderr,"signal\n");
}

int progress_func(void * data, int percent, double discr) { // progress / cancel function
   
  fprintf(stderr,"iter=%d discrepancy=%lf\n", percent, discr); fflush(stderr);
  return ibreak;

}


int main(int argc, char ** argv) {

 signal(SIGINT,ctrl_c);
 
  double positions[3], J[3], max_err, std_err;
 
  double depth   = 2.;
  double I       = 65.6; 
  double D       = -12.27;
  int background = -1;
  double mag_ampl = 0.;
  double Az = 90.;
  char target_name[20];
  int is_shift = 0;
  double nx, ny, nz;
  int is_induced = 0;
  int deriv_type = 3;
  char fixed_name[80];
  double separation = 1;
  int is_finite = 0;
  int is_custom_gradient = 0;
  fixed_name[0] = 0;
  CMagData * pFixed = NULL;
  double moment = 100;
  int is_initail_moment = 0;

  FIELDTYPE ftype = totalmagfield;

  int n_iter = 10000;

  nx = ny = 0.; nz = 1;

 /* FILE * dat = fopen("32-12.dat", "rt");
  if(!dat) {
	  fprintf(stderr,"Can not open input file\n");
	  exit(0);
  }


#define stdin dat */

  strcpy(target_name, "1");


 
  if(argc == 1) {
    fprintf(stderr,"\nUsage: %s -I val -D val -Bg val -Z val < input > output\n",argv[0]); 
    fprintf(stderr,"-I  number  -  Inclination, degrees\n"); 
    fprintf(stderr,"-D  number  -  Declination, degrees\n"); 
    fprintf(stderr,"-Z  number  - initial depth below sea bottom, m [2 m]\n"); 
    fprintf(stderr,"-Bg number  - background order (0, 1 or 2) [0]\n"); 
    fprintf(stderr,"-Tg number  - target name"); 
    fprintf(stderr,"-S          - shift all profiles except first\n");  
    fprintf(stderr,"-G nx ny nz - use gradient directed as nx, ny, nz\n");  
    fprintf(stderr,"-Ind        - assume induced mag. only\n");  
    fprintf(stderr,"-Deriv number - DNSL derivative type\n");  
    fprintf(stderr,"-Finite number - finite separation for gradient\n");  
    fprintf(stderr,"-AbsGrad       - compute for absolute gradient value\n");
    fprintf(stderr,"-AbsFiniteGrad number- compute for absolute finite grad (only along x,y,z)\n");   
    fprintf(stderr,"-Az number     - local X axis azimuth\n");
    fprintf(stderr,"-CustomGradient - input file has gradient direction defined in every point\n");
    fprintf(stderr,"-FixedLine name - name of fixed profile (see -S option)\n");  
    fprintf(stderr,"-niter number - max number of iterations. Incl. num derivatives [10000]\n");
    fprintf(stderr,"-M number - Initial value for Z moment (cgs). May be needed for abs. gradient\n");        
    exit(0);
  }


for(int i=1; i<argc; i++) {
    if(argv[i][0]=='-') {
      if(strcmp(argv[i],"-I")==0) {
	  I = atof(argv[i+1]); i++; continue;
      }
      if(strcmp(argv[i],"-D")==0) {
 	  D = atof(argv[i+1]); i++;  continue;
      }
      if(strcmp(argv[i],"-Z")==0) {
	  depth = atof(argv[i+1]); i++; continue;
      }
      if(strcmp(argv[i],"-Bg")==0) {
	  background = atoi(argv[i+1]); i++; continue;
      }
      if(strcmp(argv[i],"-Tg")==0) {
 	  strcpy(target_name, argv[i+1]); i++; continue;
      }
	  if(strcmp(argv[i],"-S")==0) {
	  is_shift = 1; continue;	  
      }
      if(strcmp(argv[i],"-G")==0) {
		nx = atof(argv[i+1]); i++; 
		ny = atof(argv[i+1]); i++; 
		nz = atof(argv[i+1]); i++;
		ftype = totalmaggradient; 
		continue;
      }
      if(strcmp(argv[i],"-Deriv")==0) {
	  deriv_type = atoi(argv[i+1]); i++; continue;
      }
      if(strcmp(argv[i],"-Ind")==0) {
	  is_induced = 1; continue;	  
      }
      if(strcmp(argv[i],"-Finite")==0) {
	  is_finite =  1;
	  separation = atof(argv[i+1]); i++; continue;
       }
      if(strcmp(argv[i],"-AbsGrad")==0) {
	  ftype = absmaggradient; continue;
       }
      if(strcmp(argv[i],"-AbsFiniteGrad")==0) {
      	 ftype = absfinitegradient; 
	     separation = atof(argv[i+1]); i++; continue;
       }      
      if(strcmp(argv[i],"-Az")==0) {
	  Az = atof(argv[i+1]); i++; continue;
       }   
      if(strcmp(argv[i],"-CustomGradient")==0) {
	  is_custom_gradient = 1;
       } 
     if(strcmp(argv[i],"-FixedLine")==0) {
	 strcpy(fixed_name, argv[i+1]); i++; continue;
       } 
     if(strcmp(argv[i],"-niter")==0) {
	     n_iter = atoi(argv[i+1]); i++; continue;
       }
     if(strcmp(argv[i],"-M")==0) {
	     moment = atof(argv[i+1])/10.; is_initail_moment = 1; i++; continue;
       }     
      }
  }


  if(is_finite) ftype =  totalfinitemaggrad;


  if(deriv_type < 1 || deriv_type > 3) deriv_type =1;

// create objects we need in inversion

 CMagDataCollection DataCollection;
 CMagCollection     ObjectCollection;

// read data
  
  char buffer[1024], date_txt[20], time_txt[20], name[20];
  const char *IFS = " \t\n";
  double x, y, z, alt, T;

  CGradProfile * prof = NULL;
  char name_old[100];
  name_old[0] = 0;

     
  fgets(buffer, 1024, stdin);
  double x_middle, y_middle;            

 
  x_middle = y_middle = 0.;            
  int k =0;      
  double T_max = 0.;
  

  while(!feof(stdin)) {
      if(!fgets(buffer, 1024, stdin)) break;
      if(buffer[0] == '#') continue;
      
      char * p  = strtok (buffer, IFS);
      if(!p) continue;

      x   =  atof(p); p = strtok(NULL, IFS);
      y   =  atof(p); p = strtok(NULL, IFS);
      z   =  atof(p); p = strtok(NULL, IFS);         
      alt =  atof(p); p = strtok(NULL, IFS);
      T   =  atof(p); p = strtok(NULL, IFS);                         

      if(is_custom_gradient) {
	  nx = atof(p); p = strtok(NULL, IFS); 
	  ny = atof(p); p = strtok(NULL, IFS); 
	  nz = atof(p); p = strtok(NULL, IFS); 
      }

      x_middle += x;
      y_middle += y;
      k++;

      sscanf(p,"%s",name);

      if(strcmp(name_old, name) && prof) {
	  DataCollection.AddData(prof);
	  ObjectCollection.AddMagObject(prof);
	  prof->ComputeLocalCoordinates();
	  prof->SetFiniteSeparation(separation); 
	  prof = NULL;
      }

       if(!prof) {
   	      prof = new CGradProfile;
	      prof->InitProfile(name);
	      if(strcmp(name, fixed_name) == 0) pFixed = prof;
	      prof->SetFieldType(ftype);
	      if(is_custom_gradient)
		  prof->SetVariableDirection();
	      else 
		  prof->SetFixedDirection(nx,ny,nz);
	      //prof->FixLinearParam(0, FALSE, 0.);
	      //prof->FixLinearParam(1, FALSE, 0.);
       }

    //z = 0.;
    if(is_custom_gradient)    
	prof->Add(x, y, 0., T, z);
    else
	prof->Add(x, y, 0., T, z, nx, ny, nz);

    T_max = MAX(T, T_max);
    strcpy(name_old, name);
   }   

    x_middle /= k;
    y_middle /= k;

    DataCollection.AddData(prof);
    prof->ComputeLocalCoordinates();
    //prof->FixLinearParam(0, FALSE, 0.);
    //prof->FixLinearParam(1, FALSE, 0.);
    prof->SetFiniteSeparation(separation); 
    ObjectCollection.AddMagObject(prof);

    double params[3];

    params[0] = x_middle;
    params[1] = y_middle;
    params[2] = depth;
    
    //params[0] = 22.245;
    //params[1] = 14.111;
    //params[2] = 0.726;

    CMagDipole * pDipole = new CMagDipole(is_induced);
  
    if(!pDipole) {
      fprintf(stderr,"Can not create pDipole. Stop.\n");
      exit(0);
    }

  // no analytical derivatives fot finite abs gradinet so far
  if(ftype == absfinitegradient) pDipole->SetObjectNonLinearAbilities(1);
  
  pDipole->SetNonLinearParams(params, 3);
  
  //pDipole->FixNonLinearParam(0, TRUE, 0);
  //pDipole->FixNonLinearParam(1, TRUE, 0);
  //pDipole->FixNonLinearParam(2, TRUE, 0.5);
 
 
  CMagObject *  pBackground = NULL;
  
  switch(background) {
    case 0:
         pBackground = new CMagBackGround; 
	 if(pBackground)  {
	     pBackground->FixLinearParam(X_OFFSET, TRUE, 0.);	 
	     pBackground->FixLinearParam(Y_OFFSET, TRUE, 0.);	 
	 }
	 break;

    case 1: 
         pBackground = new CMagBackGround; 
         break;
    case 2: pBackground = new CMagBackGround2; break;
  }

   if(pBackground)  pBackground->FixLinearParam(Z_OFFSET, TRUE, 0.);

   
   ObjectCollection.AddMagObject(pDipole);

   if(pBackground) ObjectCollection.AddMagObject(pBackground); 

   CMagInv inversion;
   inversion.SetProgressFunc(progress_func);

   inversion.SetMagData(&DataCollection);
   inversion.SetMagCollection(&ObjectCollection);
   inversion.SetEarthFieldDirections(I,D,Az);

   //CMagObject *pObj =  ObjectCollection.GetObjectAt("5031114");
   //if(pObj) {
   //   pObj->FixLinearParam(0, FALSE, 0.);
   //}

   double fit;
   Z29RLS_PARAMS rls_params;
   rls_params.dm = 1.e+15;
  

   //inversion.EstimateLinearParameters(Estimate_L2_rls, &fit, 1, &rls_params);
   //inversion.DumpObjects(stderr);
   //DataCollection.Dump(stdout);
   //exit(0);

    // add profile position to the varaibles
   
   if(is_shift) {  
      
       // find line closest to central point
 
       double x1, y1, x2, y2, z1, z2;
       DataCollection.ResetDataPos();
       CMagData * pData = DataCollection.GetNextData();
       double dist = 1.e+38;
       double dist1;

       if(!pFixed) {
       while(pData) {
	      pData->GetBoundingBox(&x1, &x2, &y1, &y2, &z1, &z2);
	      dist1 = sqrt( pow( (x1+x2)/2.-x_middle, 2) + pow( (y1+y2)/2.-y_middle, 2));
	      if(dist1 < dist) {
		  pFixed = pData;
		  dist = dist1;
		  pFixed->GetObjectName(fixed_name, 80);
	      }
	   pData = DataCollection.GetNextData();
         }
       }
       
       DataCollection.ResetDataPos();
       pData = DataCollection.GetNextData();

       while(pData && pFixed) {
	   if(pData != pFixed) {
	       pData->FixNonLinearParam(X_OFFSET, FALSE, 0.);
	       pData->FixNonLinearParam(Y_OFFSET, FALSE, 0.);
	       //pData->FixNonLinearParam(Z_OFFSET, FALSE, 0.);
	       //pData->FixLinearParam(0, FALSE, 0.);
	       //pData->FixLinearParam(1, FALSE, 0.);
	       ////ObjectCollection.AddMagObject(pData);
	   }
	   pData = DataCollection.GetNextData();
       }
   }  

   if(is_initail_moment) { // skip linear part if initial moment is set
      double jparams[3];
      jparams[0] = jparams[1] = 0;
      jparams[2] = moment;
      pDipole->SetLinearParams(jparams, 3, 0);                       
      }
   else {            
     inversion.EstimateLinearParameters(Estimate_L2_rls, &fit, 1, &rls_params);
   }

   DNLS1_STORAGE dstorage;
   dstorage.iopt =  deriv_type;
   dstorage.maxfev =  n_iter;
   char message[1024];
   inversion.EstimateAllParameters(Estimator_dnls1, &dstorage, message);

   inversion.DumpObjects(stderr);
   fprintf(stderr,"%s\n", message);

   double A, B, C, Jtotal, incl, decl, angle, dx[3], x_pos[3];
   inversion.GetEarthCoeff(&A, &B, &C, &Az);
   pDipole->GetDipoleMoment(A, B, C, Az, &Jtotal, &incl, &decl, &angle);
   pDipole->GetLinearParams(params,3);
   pDipole->GetLinearParams(J, 3);

   pDipole->GetNonLinearParams(dx, 3, 1);  
   pDipole->GetNonLinearParams(x_pos, 3);  

   fit = inversion.GetNonLinearDiscrepancy();

   fprintf(stdout,"# NAME FIT(nT/m) FIT(%%) X(m) Y(m) Z(m) Jt(cgs) INCL(deg) DECL(deg) ANGL(deg) Jx Jy Jz DX DY DZ\n");
   fprintf(stdout,"#POS %s %.2lf %.2lf  %.2lf %.2lf %.2lf   %.2lf  %.2lf %.2lf %.2lf  %lf %lf %lf   %lf %lf %lf\n",
	   target_name, fit, 100.*fit/T_max, x_pos[0], x_pos[1], x_pos[2], 10*Jtotal, incl, decl, angle, 
	   10*J[0], 10*J[1], 10*J[2], dx[0], dx[1], dx[2]);

   DataCollection.Dump(stdout);

   if(is_shift) fprintf(stderr,"Fixed line %s\n", fixed_name);

}
